/*
 * config.c: Configuration file handling
 *
 * See the main source file 'vdr.c' for copyright information and
 * how to reach the author.
 *
 * $Id: config.c 1.138 2005/09/09 15:08:59 kls Exp $
 */

#include "config.h"
#include <ctype.h>
#include <stdlib.h>
#include "device.h"
#include "i18n.h"
#include "interface.h"
#include "plugin.h"
#include "recording.h"
#include "sources.h"

// IMPORTANT NOTE: in the 'sscanf()' calls there is a blank after the '%d'
// format characters in order to allow any number of blanks after a numeric
// value!

// --- cCommand -------------------------------------------------------------

char *cCommand::result = NULL;

cCommand::cCommand(void)
{
  title = command = NULL;
  confirm = false;
  nIndent = 0;
  childs = NULL;
}

cCommand::~cCommand()
{
  free(title);
  free(command);
  delete childs;
}

bool cCommand::Parse(const char *s)
{
  const char *p = strchr(s, ':');
  if (p) {
    nIndent = 0;
#ifdef CMD_SUBMENUS
    while (*s == '-')
    {
      nIndent++;
      s++;
    }
#endif // CMD_SUBMENUS
     int l = p - s;
     if (l > 0) {
        title = MALLOC(char, l + 1);
        stripspace(strn0cpy(title, s, l + 1));
        if (!isempty(title)) {
           int l = strlen(title);
           if (l > 1 && title[l - 1] == '?') {
              confirm = true;
              title[l - 1] = 0;
              }
           command = stripspace(strdup(skipspace(p + 1)));
           return !isempty(command);
           }
        }
     }
  return false;
}

const char *cCommand::Execute(const char *Parameters)
{
  free(result);
  result = NULL;
  char *cmdbuf = NULL;
  if (Parameters)
     asprintf(&cmdbuf, "%s %s", command, Parameters);
  const char *cmd = cmdbuf ? cmdbuf : command;
  dsyslog("executing command '%s'", cmd);
  FILE *p = popen(cmd, "r");
  if (p) {
     int l = 0;
     int c;
     while ((c = fgetc(p)) != EOF) {
           if (l % 20 == 0)
              result = (char *)realloc(result, l + 21);
           result[l++] = c;
           }
     if (result)
        result[l] = 0;
     pclose(p);
     }
  else
     esyslog("ERROR: can't open pipe for command '%s'", cmd);
  free(cmdbuf);
  return result;
}

int cCommand::getIndent ()
{
  return nIndent;
}

void cCommand::setIndent (int nNewIndent)
{
  nIndent = nNewIndent;
}

bool cCommand::hasChilds ()
{
  if (!childs)
  {
    return false;
  }
  return (childs->Count () > 0);
}

int cCommand::getChildCount ()
{
  if (!childs)
  {
    return false;
  }
  return childs->Count ();
}

void cCommand::addChild (cCommand *newChild)
{
  if (!childs)
  {
    childs = new cCommands ();
  }
  childs->Add (newChild);
}


cCommands *cCommand::getChilds ()
{
  return childs;
}

// --- cCommands -------------------------------------------------------

void cCommands::AddConfig(cCommand *Object)
{
  cCommand *c = (cCommand *) Object;
  cCommand *cParent;
  int nIndent;
  int nIndex;

  if (!c)
  {
    return;
  }
  nIndent = c->getIndent ();
  //  isyslog ("nIndent %d %s\n", nIndent, c->Title ());
  for (nIndex = Count () - 1; nIndex >= 0; nIndex--)
  {
    cParent = (cCommand *) Get (nIndex);
    if (cParent->getIndent () < nIndent)
    {
      cParent->addChild (c);
      return;
    }
  }
  cConfig<cCommand>::Add(Object);
}

// -- cSVDRPhost -------------------------------------------------------------

cSVDRPhost::cSVDRPhost(void)
{
  addr.s_addr = 0;
  mask = 0;
}

bool cSVDRPhost::Parse(const char *s)
{
  mask = 0xFFFFFFFF;
  const char *p = strchr(s, '/');
  if (p) {
     char *error = NULL;
     int m = strtoul(p + 1, &error, 10);
     if (error && *error && !isspace(*error) || m > 32)
        return false;
     *(char *)p = 0; // yes, we know it's 'const' - will be restored!
     if (m == 0)
        mask = 0;
     else
        mask >>= (32 - m);
     }
  int result = inet_aton(s, &addr);
  if (p)
     *(char *)p = '/'; // there it is again
  return result != 0 && (mask != 0 || addr.s_addr == 0);
}

bool cSVDRPhost::Accepts(in_addr_t Address)
{
  return (Address & mask) == addr.s_addr;
}

// -- cCaDefinition ----------------------------------------------------------

cCaDefinition::cCaDefinition(void)
{
  number = 0;
  description = NULL;
}

cCaDefinition::~cCaDefinition()
{
  free(description);
}

bool cCaDefinition::Parse(const char *s)
{
  return 2 == sscanf(s, "%d %a[^\n]", &number, &description) && description && *description;
}

// -- cCommands --------------------------------------------------------------

cCommands Commands;
cCommands RecordingCommands;
cCommands TimerCommands;

// -- cSVDRPhosts ------------------------------------------------------------

cSVDRPhosts SVDRPhosts;

bool cSVDRPhosts::Acceptable(in_addr_t Address)
{
  cSVDRPhost *h = First();
  while (h) {
        if (h->Accepts(Address))
           return true;
        h = (cSVDRPhost *)h->Next();
        }
  return false;
}

// -- cCaDefinitions ---------------------------------------------------------

cCaDefinitions CaDefinitions;

const cCaDefinition *cCaDefinitions::Get(int Number)
{
  cCaDefinition *p = First();
  while (p) {
        if (p->Number() == Number)
           return p;
        p = (cCaDefinition *)p->Next();
        }
  return NULL;
}

// -- cSetupLine -------------------------------------------------------------

cSetupLine::cSetupLine(void)
{
  plugin = name = value = NULL;
}

cSetupLine::cSetupLine(const char *Name, const char *Value, const char *Plugin)
{
  name = strdup(Name);
  value = strdup(Value);
  plugin = Plugin ? strdup(Plugin) : NULL;
}

cSetupLine::~cSetupLine()
{
  free(plugin);
  free(name);
  free(value);
}

int cSetupLine::Compare(const cListObject &ListObject) const
{
  const cSetupLine *sl = (cSetupLine *)&ListObject;
  if (!plugin && !sl->plugin)
     return strcasecmp(name, sl->name);
  if (!plugin)
     return -1;
  if (!sl->plugin)
     return 1;
  int result = strcasecmp(plugin, sl->plugin);
  if (result == 0)
     result = strcasecmp(name, sl->name);
  return result;
}

bool cSetupLine::Parse(char *s)
{
  char *p = strchr(s, '=');
  if (p) {
     *p = 0;
     char *Name  = compactspace(s);
     char *Value = compactspace(p + 1);
     if (*Name) { // value may be an empty string
        p = strchr(Name, '.');
        if (p) {
           *p = 0;
           char *Plugin = compactspace(Name);
           Name = compactspace(p + 1);
           if (!(*Plugin && *Name))
              return false;
           plugin = strdup(Plugin);
           }
        name = strdup(Name);
        value = strdup(Value);
        return true;
        }
     }
  return false;
}

bool cSetupLine::Save(FILE *f)
{
  return fprintf(f, "%s%s%s = %s\n", plugin ? plugin : "", plugin ? "." : "", name, value) > 0;
}

// -- cSetup -----------------------------------------------------------------

cSetup Setup;

cSetup::cSetup(void)
{
  OSDLanguage = 0;
  strcpy(OSDSkin, "sttng");
  strcpy(OSDTheme, "default");
  WarEagleIcons = 1;
  PrimaryDVB = 1;
  ShowInfoOnChSwitch = 1;
  MenuScrollPage = 1;
  MenuScrollWrap = 0;
  MarkInstantRecord = 1;
  strcpy(NameInstantRecord, "TITLE EPISODE");
  InstantRecordTime = 180;
  LnbSLOF    = 11700;
  LnbFrequLo =  9750;
  LnbFrequHi = 10600;
  DiSEqC = 0;
  SetSystemTime = 0;
  TimeSource = 0;
  TimeTransponder = 0;
  MarginStart = 2;
  MarginStop = 10;
  JumpSeconds = 60;
  JumpSecondsSlow = 10;
  AudioLanguages[0] = -1;
  EPGLanguages[0] = -1;
  EPGScanTimeout = 5;
  EPGBugfixLevel = 3;
  EPGLinger = 0;
  SVDRPTimeout = 300;
  ZapTimeout = 3;
  DirectChannelListSwitch = 0;
  ShowValidInput = 0;
  SortTimers = 1;
  SortSwitchTimersSep = 1;
  ConfirmSwitchTimers = 0;
  ConfirmSwitchTimersTimeout = 30;
  BootForSwitchTimer = 0;
  PrimaryLimit = 0;
  DefaultPriority = 50;
  DefaultLifetime = 99;
  PausePriority = 10;
  PauseLifetime = 1;
  UseSubtitle = 1;
  UseVps = 0;
  LiveBuffer = 0;
  LiveBufferSize = 10;
  LiveReplay = 0;
  KeepBuffer = 0;
  KeepLastBuffer = 0;
  NumLastBuffers = 1;
  BufferTimeOut = 30;
  MinLiveBufferInactivity = 0;
  Frameswait = 0;
  VpsMargin = 120;
  RecordingDirs = 1;
  SortRecords = 2;
  SortRecordsHelper = 2;
  VideoDisplayFormat = 1;
  VideoFormat = 0;
  UpdateChannels = 5;
  UseDolbyDigital = 1;
  RecordDolbyDigital = 1;
  ChannelInfoPos = 0;
  ChannelInfoTime = 5;
  OSDLeft = 54;
  OSDTop = 45;
  OSDWidth = 624;
  OSDHeight = 486;
  OSDMessageTime = 1;
  UseSmallFont = 1;
  MaxVideoFileSize = MAXVIDEOFILESIZE;
  SplitEditedFiles = 0;
  MinEventTimeout = 30;
  MinUserInactivity = 300;
  MultiSpeedMode = 0;
  ShowReplayMode = 0;
  DoubleEpgTimeDelta = 15;
  DoubleEpgAction = 0;  
  MixEpgAction = 0;
  DisableVPS = 0;
  ResumeID = 0;
  memset(SourceCaps, 0, sizeof SourceCaps);
  SourceCapsSet = false;
  JumpPlay = 0;
  PlayJump = 0;
  LoadMarksInterval = 0;
  CurrentChannel = -1;
//ML
  VerboseLNBlog = 0;
  for (int i = 0; i < MAXDEVICES; i++) CardUsesLNBnr[i] = i + 1;
//ML-Ende
  CurrentVolume = MAXVOLUME;
  CurrentDolby = 0;
  strcpy(noEPG, "");
  noEPGPatch = 0;
  LocalChannelProvide = 1;
  PreferredChannel = 0; // Save the current channel to setup.conf
  StartupVolume = 256;
#ifdef DVDARCHIVE
  TrayOpenOnBegin = 0;
  TrayOpenOnEnd = 0;
#endif /* DVDARCHIVE */
  ShowRecDate = 0;
  ShowRecTime = 0;
  ShowRecLength = 1;
  ShowRecType = 1;
  ShowRecPathLength = 25;
  MenuCmdPosition = 0;
  ShowProgressBar = 0;
  ShowTimerInfo = 0;
}

cSetup& cSetup::operator= (const cSetup &s)
{
  memcpy(&__BeginData__, &s.__BeginData__, (char *)&s.__EndData__ - (char *)&s.__BeginData__);
  return *this;
}

cSetupLine *cSetup::Get(const char *Name, const char *Plugin)
{
  for (cSetupLine *l = First(); l; l = Next(l)) {
      if ((l->Plugin() == NULL) == (Plugin == NULL)) {
         if ((!Plugin || strcasecmp(l->Plugin(), Plugin) == 0) && strcasecmp(l->Name(), Name) == 0)
            return l;
         }
      }
  return NULL;
}

void cSetup::Store(const char *Name, const char *Value, const char *Plugin, bool AllowMultiple)
{
  if (Name && *Name) {
     cSetupLine *l = Get(Name, Plugin);
     if (l && !AllowMultiple)
        Del(l);
     if (Value)
        Add(new cSetupLine(Name, Value, Plugin));
     }
}

void cSetup::Store(const char *Name, int Value, const char *Plugin)
{
  char *buffer = NULL;
  asprintf(&buffer, "%d", Value);
  Store(Name, buffer, Plugin);
  free(buffer);
}

bool cSetup::Load(const char *FileName)
{
  if (cConfig<cSetupLine>::Load(FileName, true)) {
     bool result = true;
     for (cSetupLine *l = First(); l; l = Next(l)) {
         bool error = false;
         if (l->Plugin()) {
            cPlugin *p = cPluginManager::GetPlugin(l->Plugin());
            if (p && !p->SetupParse(l->Name(), l->Value()))
               error = true;
            }
         else {
            if (!Parse(l->Name(), l->Value()))
               error = true;
            }
         if (error) {
//            esyslog("ERROR: unknown config parameter: %s%s%s = %s", l->Plugin() ? l->Plugin() : "", l->Plugin() ? "." : "", l->Name(), l->Value());
            result = false;
            }
         }
     return result;
     }
  return false;
}

void cSetup::StoreSourceCaps(const char *Name)
{
  cSetupLine *l;
  while ((l = Get(Name)) != NULL)
        Del(l);

  for (int i = 0; i < MAXDEVICES; i++) {
      char buffer[MAXPARSEBUFFER]={0,}, *q = buffer;
      int j = 0;
      while (SourceCaps[i][j] && j < MAXSOURCECAPS) {
            if (j==0) 
               q += snprintf(buffer, sizeof(buffer), "%i ", i+1);
            q += snprintf(q, sizeof(buffer) - (q-buffer), "%s ", *cSource::ToString(SourceCaps[i][j++]));
      }
      if (*buffer)
         Store(Name, buffer, NULL, true);
      }
}

bool cSetup::ParseSourceCaps(const char *Value)
{
  char *p;
  int d = strtol(Value, &p, 10)-1, i = 0;
  while (p < Value+strlen(Value)) {
     	if (*p==0) return true;
     	if (isblank(*p)) ++p;
     	if (isalpha(*p)) {
	   int source = cSource::FromString(p);
	   if (source != cSource::stNone) {
	      SourceCaps[d][i++] = source;
	      SourceCapsSet = true;
	      }
	   else
	      return false;
	   while (!isblank(*p) && *p)
		 ++p;
	   if (i>MAXSOURCECAPS) 
	      return false;
     	   }
        }
  return true;
}

void cSetup::StoreLanguages(const char *Name, int *Values)
{
  char buffer[I18nNumLanguages * 4];
  char *q = buffer;
  for (int i = 0; i < I18nNumLanguages; i++) {
      if (Values[i] < 0)
         break;
      const char *s = I18nLanguageCode(Values[i]);
      if (s) {
         if (q > buffer)
            *q++ = ' ';
         strncpy(q, s, 3);
         q += 3;
         }
      }
  *q = 0;
  Store(Name, buffer);
}

bool cSetup::ParseLanguages(const char *Value, int *Values)
{
  int n = 0;
  while (Value && *Value && n < I18nNumLanguages) {
        char buffer[4];
        strn0cpy(buffer, Value, sizeof(buffer));
        int i = I18nLanguageIndex(buffer);
        if (i >= 0)
           Values[n++] = i;
        if ((Value = strchr(Value, ' ')) != NULL)
           Value++;
        }
  Values[n] = -1;
  return true;
}

bool cSetup::Parse(const char *Name, const char *Value)
{
  if      (!strcasecmp(Name, "OSDLanguage"))         OSDLanguage        = atoi(Value);
  else if (!strcasecmp(Name, "OSDSkin"))             strn0cpy(OSDSkin, Value, MaxSkinName);
  else if (!strcasecmp(Name, "OSDTheme"))            strn0cpy(OSDTheme, Value, MaxThemeName);
  else if (!strcasecmp(Name, "WarEagleIcons"))       WarEagleIcons      = atoi(Value);
  else if (!strcasecmp(Name, "PrimaryDVB"))          PrimaryDVB         = atoi(Value);
  else if (!strcasecmp(Name, "ShowInfoOnChSwitch"))  ShowInfoOnChSwitch = atoi(Value);
  else if (!strcasecmp(Name, "MenuScrollPage"))      MenuScrollPage     = atoi(Value);
  else if (!strcasecmp(Name, "MenuScrollWrap"))      MenuScrollWrap     = atoi(Value);
  else if (!strcasecmp(Name, "MarkInstantRecord"))   MarkInstantRecord  = atoi(Value);
  else if (!strcasecmp(Name, "NameInstantRecord"))   strn0cpy(NameInstantRecord, Value, MaxFileName);
  else if (!strcasecmp(Name, "InstantRecordTime"))   InstantRecordTime  = atoi(Value);
  else if (!strcasecmp(Name, "LnbSLOF"))             LnbSLOF            = atoi(Value);
  else if (!strcasecmp(Name, "LnbFrequLo"))          LnbFrequLo         = atoi(Value);
  else if (!strcasecmp(Name, "LnbFrequHi"))          LnbFrequHi         = atoi(Value);
  else if (!strcasecmp(Name, "DiSEqC"))              DiSEqC             = atoi(Value);
  else if (!strcasecmp(Name, "SetSystemTime"))       SetSystemTime      = atoi(Value);
  else if (!strcasecmp(Name, "TimeSource"))          TimeSource         = cSource::FromString(Value);
  else if (!strcasecmp(Name, "TimeTransponder"))     TimeTransponder    = atoi(Value);
  else if (!strcasecmp(Name, "MarginStart"))         MarginStart        = atoi(Value);
  else if (!strcasecmp(Name, "MarginStop"))          MarginStop         = atoi(Value);
  else if (!strcasecmp(Name, "JumpSeconds"))		 JumpSeconds		= atoi(Value);
  else if (!strcasecmp(Name, "JumpSecondsSlow"))	 JumpSecondsSlow	= atoi(Value);
  else if (!strcasecmp(Name, "AudioLanguages"))      return ParseLanguages(Value, AudioLanguages);
  else if (!strcasecmp(Name, "EPGLanguages"))        return ParseLanguages(Value, EPGLanguages);
  else if (!strcasecmp(Name, "EPGScanTimeout"))      EPGScanTimeout     = atoi(Value);
  else if (!strcasecmp(Name, "EPGBugfixLevel"))      EPGBugfixLevel     = atoi(Value);
  else if (!strcasecmp(Name, "EPGLinger"))           EPGLinger          = atoi(Value);
  else if (!strcasecmp(Name, "SVDRPTimeout"))        SVDRPTimeout       = atoi(Value);
  else if (!strcasecmp(Name, "ZapTimeout"))          ZapTimeout         = atoi(Value);
  else if (!strcasecmp(Name, "DirectChannelListSwitch")) DirectChannelListSwitch = atoi(Value);
  else if (!strcasecmp(Name, "ShowValidInput"))      ShowValidInput     = atoi(Value);
  else if (!strcasecmp(Name, "SortTimers"))          SortTimers         = atoi(Value);
  else if (!strcasecmp(Name, "SortSwitchTimersSep")) SortSwitchTimersSep= atoi(Value);
  else if (!strcasecmp(Name, "ConfirmSwitchTimers")) ConfirmSwitchTimers= atoi(Value);
  else if (!strcasecmp(Name, "ConfirmSwitchTimersTimeout")) ConfirmSwitchTimersTimeout= atoi(Value);
  else if (!strcasecmp(Name, "BootForSwitchTimer"))  BootForSwitchTimer = atoi(Value);
  else if (!strcasecmp(Name, "PrimaryLimit"))        PrimaryLimit       = atoi(Value);
  else if (!strcasecmp(Name, "DefaultPriority"))     DefaultPriority    = atoi(Value);
  else if (!strcasecmp(Name, "DefaultLifetime"))     DefaultLifetime    = atoi(Value);
  else if (!strcasecmp(Name, "PausePriority"))       PausePriority      = atoi(Value);
  else if (!strcasecmp(Name, "PauseLifetime"))       PauseLifetime      = atoi(Value);
  else if (!strcasecmp(Name, "UseSubtitle"))         UseSubtitle        = atoi(Value);
  else if (!strcasecmp(Name, "UseVps"))              UseVps             = atoi(Value);
  else if (!strcasecmp(Name, "VpsMargin"))           VpsMargin          = atoi(Value);
  else if (!strcasecmp(Name, "LiveBuffer"))	         LiveBuffer         = atoi(Value);
  else if (!strcasecmp(Name, "LiveBufferSize"))      LiveBufferSize     = atoi(Value);
  else if (!strcasecmp(Name, "LiveReplay"))          LiveReplay         = atoi(Value);
  else if (!strcasecmp(Name, "KeepBuffer"))          KeepBuffer         = atoi(Value);
  else if (!strcasecmp(Name, "KeepLastBuffer"))      KeepLastBuffer     = atoi(Value);
  else if (!strcasecmp(Name, "NumLastBuffers"))      NumLastBuffers     = atoi(Value);
  else if (!strcasecmp(Name, "BufferTimeOut"))       BufferTimeOut      = atoi(Value);  
  else if (!strcasecmp(Name, "MinLiveBufferInactivity")) MinLiveBufferInactivity = atoi(Value);
  else if (!strcasecmp(Name, "Frameswait"))          Frameswait         = atoi(Value);
  else if (!strcasecmp(Name, "RecordingDirs"))       RecordingDirs      = atoi(Value);
  else if (!strcasecmp(Name, "VideoDisplayFormat"))  VideoDisplayFormat = atoi(Value);
  else if (!strcasecmp(Name, "VideoFormat"))         VideoFormat        = atoi(Value);
  else if (!strcasecmp(Name, "UpdateChannels"))      UpdateChannels     = atoi(Value);
  else if (!strcasecmp(Name, "UseDolbyDigital"))     UseDolbyDigital    = atoi(Value);
  else if (!strcasecmp(Name, "RecordDolbyDigital"))  RecordDolbyDigital = atoi(Value);
  else if (!strcasecmp(Name, "ChannelInfoPos"))      ChannelInfoPos     = atoi(Value);
  else if (!strcasecmp(Name, "ChannelInfoTime"))     ChannelInfoTime    = atoi(Value);
  else if (!strcasecmp(Name, "OSDLeft"))             OSDLeft            = atoi(Value);
  else if (!strcasecmp(Name, "OSDTop"))              OSDTop             = atoi(Value);
  else if (!strcasecmp(Name, "OSDWidth"))          { OSDWidth           = atoi(Value); if (OSDWidth  < 100) OSDWidth  *= 12; OSDWidth &= ~0x07; } // OSD width must be a multiple of 8
  else if (!strcasecmp(Name, "OSDHeight"))         { OSDHeight          = atoi(Value); if (OSDHeight < 100) OSDHeight *= 27; }
  else if (!strcasecmp(Name, "OSDMessageTime"))      OSDMessageTime     = atoi(Value);
  else if (!strcasecmp(Name, "UseSmallFont"))        UseSmallFont       = atoi(Value);
  else if (!strcasecmp(Name, "MaxVideoFileSize"))    MaxVideoFileSize   = atoi(Value);
  else if (!strcasecmp(Name, "SplitEditedFiles"))    SplitEditedFiles   = atoi(Value);
  else if (!strcasecmp(Name, "MinEventTimeout"))     MinEventTimeout    = atoi(Value);
  else if (!strcasecmp(Name, "MinUserInactivity"))   MinUserInactivity  = atoi(Value);
  else if (!strcasecmp(Name, "MultiSpeedMode"))      MultiSpeedMode     = atoi(Value);
  else if (!strcasecmp(Name, "ShowReplayMode"))      ShowReplayMode     = atoi(Value);
  else if (!strcasecmp(Name, "noEPG"))               strn0cpy(noEPG, Value, 999);
  else if (!strcasecmp(Name, "noEPGPatch"))          noEPGPatch         = atoi(Value);
  else if (!strcasecmp(Name, "DoubleEpgTimeDelta"))  DoubleEpgTimeDelta = atoi(Value);
  else if (!strcasecmp(Name, "DoubleEpgAction"))     DoubleEpgAction    = atoi(Value);
  else if (!strcasecmp(Name, "MixEpgAction"))        MixEpgAction    	= atoi(Value);
  else if (!strcasecmp(Name, "DisableVPS"))          DisableVPS     	= atoi(Value);
  else if (!strcasecmp(Name, "ResumeID"))            ResumeID           = atoi(Value);
  else if (!strcasecmp(Name, "SourceCaps"))          return    ParseSourceCaps(Value);
  else if (!strcasecmp(Name, "JumpPlay"))            JumpPlay           = atoi(Value);
  else if (!strcasecmp(Name, "PlayJump"))            PlayJump           = atoi(Value);
  else if (!strcasecmp(Name, "LoadMarksInterval"))   LoadMarksInterval  = atoi(Value);
  else if (!strcasecmp(Name, "CurrentChannel"))      CurrentChannel     = atoi(Value);
  else if (!strcasecmp(Name, "CurrentVolume"))       CurrentVolume      = atoi(Value);
  else if (!strcasecmp(Name, "CurrentDolby"))        CurrentDolby       = atoi(Value);
  else if (!strcasecmp(Name, "SortRecords"))         SortRecords        = atoi(Value);
  else if (!strcasecmp(Name, "SortRecordsHelper"))   SortRecordsHelper  = atoi(Value);
  else if (!strcasecmp(Name, "LocalChannelProvide")) LocalChannelProvide= atoi(Value);
  else if (!strcasecmp(Name, "PreferredChannel")) {
     PreferredChannel      = atoi(Value);
     if (PreferredChannel > 0) CurrentChannel = PreferredChannel;
   }
   else if (!strcasecmp(Name, "StartupVolume")) {
     StartupVolume      = atoi(Value);
     if (StartupVolume < 256) CurrentVolume = StartupVolume;
   }
 
#ifdef DVDARCHIVE
  else if (!strcasecmp(Name, "TrayOpenOnBegin"))     TrayOpenOnBegin    = atoi(Value);
  else if (!strcasecmp(Name, "TrayOpenOnEnd"))       TrayOpenOnEnd      = atoi(Value);
#endif /* DVDARCHIVE */

  else if (!strcasecmp(Name, "ShowRecDate"))         ShowRecDate        = atoi(Value);
  else if (!strcasecmp(Name, "ShowRecTime"))         ShowRecTime        = atoi(Value);
  else if (!strcasecmp(Name, "ShowRecLength"))       ShowRecLength      = atoi(Value);
  else if (!strcasecmp(Name, "ShowRecType"))         ShowRecType        = atoi(Value);
  else if (!strcasecmp(Name, "ShowRecPathLength"))   ShowRecPathLength  = atoi(Value);
  else if (!strcasecmp(Name, "MenuCmdPosition"))     MenuCmdPosition    = atoi(Value);
  else if (!strcasecmp(Name, "ShowProgressBar"))     ShowProgressBar    = atoi(Value);
  else if (!strcasecmp(Name, "ShowTimerInfo"))       ShowTimerInfo      = atoi(Value);
  else

//ML
  if (!strcasecmp(Name, "VerboseLNBlog")) VerboseLNBlog = atoi(Value);
  else {
    char tmp[20];
    bool result = false;
    for (int i = 1; i <= MAXDEVICES; i++) {
      sprintf(tmp, "Card%dusesLNBnr", i);
      if (!strcasecmp(Name, tmp)) {
        CardUsesLNBnr[i - 1] = atoi(Value);
        result = true;
      }
    }  
     return result;
  }
//ML-Ende

  return true;
}

bool cSetup::Save(void)
{
  Store("OSDLanguage",        OSDLanguage);
  Store("OSDSkin",            OSDSkin);
  Store("OSDTheme",           OSDTheme);
  Store("WarEagleIcons",      WarEagleIcons);
  Store("PrimaryDVB",         PrimaryDVB);
  Store("ShowInfoOnChSwitch", ShowInfoOnChSwitch);
  Store("MenuScrollPage",     MenuScrollPage);
  Store("MenuScrollWrap",     MenuScrollWrap);
  Store("MarkInstantRecord",  MarkInstantRecord);
  Store("NameInstantRecord",  NameInstantRecord);
  Store("InstantRecordTime",  InstantRecordTime);
  Store("LnbSLOF",            LnbSLOF);
  Store("LnbFrequLo",         LnbFrequLo);
  Store("LnbFrequHi",         LnbFrequHi);
  Store("DiSEqC",             DiSEqC);
  Store("SetSystemTime",      SetSystemTime);
  Store("TimeSource",         cSource::ToString(TimeSource));
  Store("TimeTransponder",    TimeTransponder);
  Store("MarginStart",        MarginStart);
  Store("MarginStop",         MarginStop);
  Store("JumpSeconds",		  JumpSeconds);
  Store("JumpSecondsSlow",	  JumpSecondsSlow);
  StoreLanguages("AudioLanguages", AudioLanguages);
  StoreLanguages("EPGLanguages", EPGLanguages);
  Store("EPGScanTimeout",     EPGScanTimeout);
  Store("EPGBugfixLevel",     EPGBugfixLevel);
  Store("EPGLinger",          EPGLinger);
  Store("SVDRPTimeout",       SVDRPTimeout);
  Store("ZapTimeout",         ZapTimeout);
  Store("DirectChannelListSwitch", DirectChannelListSwitch);
  Store("ShowValidInput",     ShowValidInput);
  Store("SortTimers",         SortTimers);
  Store("SortSwitchTimersSep",SortSwitchTimersSep);
  Store("ConfirmSwitchTimers",ConfirmSwitchTimers);
  Store("ConfirmSwitchTimersTimeout",ConfirmSwitchTimersTimeout);
  Store("BootForSwitchTimer", BootForSwitchTimer);
  Store("PrimaryLimit",       PrimaryLimit);
  Store("DefaultPriority",    DefaultPriority);
  Store("DefaultLifetime",    DefaultLifetime);
  Store("PausePriority",      PausePriority);
  Store("PauseLifetime",      PauseLifetime);
  Store("UseSubtitle",        UseSubtitle);
  Store("UseVps",             UseVps);
  Store("VpsMargin",          VpsMargin);
  Store("LiveBuffer",         LiveBuffer);
  Store("LiveBufferSize",     LiveBufferSize);
  Store("LiveReplay",         LiveReplay);
  Store("KeepBuffer",         KeepBuffer);
  Store("KeepLastBuffer",     KeepLastBuffer);
  Store("NumLastBuffers",     NumLastBuffers);
  Store("BufferTimeOut",      BufferTimeOut);
  Store("MinLiveBufferInactivity", MinLiveBufferInactivity);
  Store("Frameswait",         Frameswait);
  Store("RecordingDirs",      RecordingDirs);
  Store("VideoDisplayFormat", VideoDisplayFormat);
  Store("VideoFormat",        VideoFormat);
  Store("UpdateChannels",     UpdateChannels);
  Store("UseDolbyDigital",    UseDolbyDigital);
  Store("RecordDolbyDigital", RecordDolbyDigital);
  Store("ChannelInfoPos",     ChannelInfoPos);
  Store("ChannelInfoTime",    ChannelInfoTime);
  Store("OSDLeft",            OSDLeft);
  Store("OSDTop",             OSDTop);
  Store("OSDWidth",           OSDWidth);
  Store("OSDHeight",          OSDHeight);
  Store("OSDMessageTime",     OSDMessageTime);
  Store("UseSmallFont",       UseSmallFont);
  Store("MaxVideoFileSize",   MaxVideoFileSize);
  Store("SplitEditedFiles",   SplitEditedFiles);
  Store("MinEventTimeout",    MinEventTimeout);
  Store("MinUserInactivity",  MinUserInactivity);
  Store("DoubleEpgAction",    DoubleEpgAction);
  Store("MixEpgAction",	      MixEpgAction);
  Store("DisableVPS",	      DisableVPS);
  Store("DoubleEpgTimeDelta", DoubleEpgTimeDelta);
  Store("MultiSpeedMode",     MultiSpeedMode);
  Store("ShowReplayMode",     ShowReplayMode);
  Store("ResumeID",           ResumeID);
  StoreSourceCaps("SourceCaps");
  Store("JumpPlay",           JumpPlay);
  Store("PlayJump",           PlayJump);
  Store("LoadMarksInterval",  LoadMarksInterval);
  Store("CurrentChannel",     PreferredChannel==0 ? CurrentChannel : PreferredChannel);
  Store("CurrentVolume",      CurrentVolume);
  Store("CurrentDolby",       CurrentDolby);
  Store("SortRecords",        SortRecords);
  Store("SortRecordsHelper",  SortRecordsHelper);
  Store("noEPG",              noEPG);
  Store("noEPGPatch",         noEPGPatch);
  Store("LocalChannelProvide",LocalChannelProvide);
#ifdef DVDARCHIVE
  Store("TrayOpenOnBegin",    TrayOpenOnBegin);
  Store("TrayOpenOnEnd",      TrayOpenOnEnd);
#endif /* DVDARCHIVE */
  Store("ShowRecDate",        ShowRecDate);
  Store("ShowRecTime",        ShowRecTime);
  Store("ShowRecLength",      ShowRecLength);
  Store("ShowRecType",        ShowRecType);
  Store("ShowRecPathLength",  ShowRecPathLength);
  Store("MenuCmdPosition",    MenuCmdPosition);
  Store("ShowProgressBar",    ShowProgressBar);
  Store("PreferredChannel",   PreferredChannel);
  Store("StartupVolume",      StartupVolume);
  Store("ShowTimerInfo",      ShowTimerInfo);
//ML
  Store("VerboseLNBlog",       VerboseLNBlog);
  char tmp[20];
  if (cDevice::NumDevices() > 1) {
     for (int i = 1; i <= cDevice::NumDevices(); i++) {
        sprintf(tmp, "Card%dusesLNBnr", i);
        Store(tmp, CardUsesLNBnr[i - 1]);
     }
  }
//ML-Ende

  Sort();

  if (cConfig<cSetupLine>::Save()) {
     isyslog("saved setup to %s", FileName());
     return true;
     }
  return false;
}
